import numpy as np
import pandas as pd
import pathlib
import logging
import random

logger = logging.getLogger(__name__)

class NN_handler():

    """
    This class contains code for neural network manipulation. 
    Neural network, Model creation or loading, batch creation, model training and hyperparameter optimisation (threshold)
    Connection to the data is made through the experiment class
    """

    def __init__(self, experiment, matrice_size=120, margin=1, table_index=0, exclude=[]):
        self.experiment = experiment
        self.annotation_table = experiment.feature_tables[table_index].annotation_table
        self.labels = self.get_labels(exclude)
        self.label_list = sorted(self.labels)
        self.matrice_size = matrice_size
        self.margin = margin


    def get_labels(self, exclude=[]):
        self.labels = [annotation.label for annotation in self.annotation_table.annotations if (annotation.label not in exclude)]  
        return self.labels


    def create_batches(self, validation_split=0.1, normalise_class=False, merge_classes=[]):
        from sklearn.preprocessing import LabelEncoder
        from keras.utils import to_categorical
        # Get all the peaks that have been annotated
        all_peaks = [peak for annotation in self.annotation_table.annotations for peak in annotation.peaks if (annotation.label in self.labels)]
        # Randomise them as they come by order of annotation
        random.shuffle(all_peaks)

        # If merge classes is on, redefine the label list
        if len(merge_classes) > 0:
            self.label_list = sorted(list(set([label if label not in merge_classes[0] else merge_classes[1] for label in self.labels])))

        # This is to get an equal number of training point (peak) for all different classes
        # If this is not used, the sets are unbalanced towards noise peaks
        if normalise_class:
            # initialise smallest class to a very high value
            smallest_class = 1000000
            if len(merge_classes) > 0:
                merged_class_size = 0
                for annotation in self.annotation_table.annotations:
                    if annotation.label in self.labels:
                        if annotation.label in merge_classes[0]:
                            merged_class_size += len(annotation.peaks)
                        else:
                            class_size = len(annotation.peaks)
                            smallest_class = class_size if class_size < smallest_class else smallest_class
                smallest_class = merged_class_size if merged_class_size < smallest_class else smallest_class
            else:
                for annotation in self.annotation_table.annotations:
                    if annotation.label in self.labels:
                        class_size = len(annotation.peaks)
                        smallest_class = class_size if class_size < smallest_class else smallest_class

            peaks = []
            class_count = [0] * len(self.label_list)
            for peak in all_peaks:
                if (len(merge_classes) > 0) and (peak.annotation.label in merge_classes[0]):
                    peak_class = merge_classes[1]
                    if class_count[self.label_list.index(merge_classes[1])] < smallest_class:
                        peaks.append(peak)
                        class_count[self.label_list.index(merge_classes[1])] += 1
                else:
                    if class_count[self.label_list.index(peak.annotation.label)] < smallest_class:
                        peaks.append(peak)
                        class_count[self.label_list.index(peak.annotation.label)] += 1

        else:
            peaks = all_peaks

        # Randomise again to avoid that one label only is stacked at the end of the list
        random.shuffle(peaks)

        # Create placeholder lists to store the data
        data = []
        labels = []
        ids = []
        num_data = []

        max_height = 0
        max_rt_window = 0
        max_mz_window = 0
        max_scan_number = 0

        # Hotfix to accelerate batch creation by extracting chromatograms using using sample module
        # Check if chromatograms do not already exist
        create_chromatograms = False
        for peak in peaks:
            if not peak.formatted_chromatogram_exists(self.matrice_size, self.margin):
                create_chromatograms = True
                break

        # Get all samples and its associated peaks present in the current peak list and store them
        if create_chromatograms:
            sample_peak_dict = dict()
            for peak in peaks: 
                if peak.sample in sample_peak_dict:
                    sample_peak_dict[peak.sample].append(peak)
                else:
                    sample_peak_dict[peak.sample] = [peak]

            for sample_key in sample_peak_dict:
                sample_key.create_interpolated_chromatograms(self.matrice_size, self.margin, peak_list=sample_peak_dict[sample_key])
        # End hotfix

        valid_peaks = []
        for peak in peaks:

            if peak.valid:
                valid_peaks.append(peak)
                # Get the interpolated, baseline corrected and normalized peak
                # dataframe = peak.interpolate_chromatogram(self.matrice_size,  self.margin, True)
                dataframe = peak.get_formatted_chromatogram(self.matrice_size, self.margin)

                # Extract only the Intensity and window value
                data.append([dataframe[0], dataframe[1]])
                # If merge classes is on, replace the classes to merge by the unique class given
                if len(merge_classes) > 0:
                    if peak.annotation.label in merge_classes[0]:
                        labels.append(merge_classes[1])
                    else:
                        labels.append(peak.annotation.label)
                else:
                    labels.append(peak.annotation.label)
                ids.append(peak.id)

        # Create an encoder and fit the label list 
        label_encoder = LabelEncoder()
        label_encoder.fit(self.label_list)

        # Find the index to create training/test split
        split_index = int(round(validation_split*len(valid_peaks)))

        # Create test data
        test_data = data[:split_index]
        test_labels = labels[:split_index]
        test_ids = ids[:split_index]

        # Create validation data
        val_data = data[split_index:split_index*2]
        val_labels = labels[split_index:split_index*2]
        val_ids = ids[split_index:split_index*2]

        # Create training data
        training_data = data[split_index*2:]
        training_labels = labels[split_index*2:]
        training_ids = ids[split_index*2:]
        

        self.test_batch = dict(
            ids = test_ids,
            data = np.asarray(test_data),
            batch_label = 'test_batch',
            labels = label_encoder.transform(np.asarray(test_labels)),
            label_str = test_labels)

        self.training_batch = dict(
            ids = training_ids,
            data = np.asarray(training_data),
            batch_label = 'training_batch',
            labels = label_encoder.transform(np.asarray(training_labels)),
            label_str  = training_labels)

        self.validation_batch = dict(
            ids = val_ids,
            data = np.asarray(val_data),
            batch_label = 'validation_batch',
            labels = label_encoder.transform(np.asarray(val_labels)),
            label_str = val_labels)

        # Format the data to feed the CNN
        self.training_batch_size = len(self.training_batch['ids'])
        self.test_batch_size = len(self.test_batch['ids'])
        self.validation_batch_size = len(self.validation_batch['ids'])

        # Reshape all batches of data
        self.X_train = self.training_batch['data'].reshape(self.training_batch_size,2,self.matrice_size,1).astype('float32')
        self.X_test = self.test_batch['data'].reshape(self.test_batch_size,2,self.matrice_size,1).astype('float32')
        self.X_validation = self.validation_batch['data'].reshape(self.validation_batch_size,2,self.matrice_size,1).astype('float32')

        # One hot encoding of labels
        self.y_train = to_categorical(self.training_batch['labels'])
        self.y_test = to_categorical(self.test_batch['labels'])
        self.y_validation = to_categorical(self.validation_batch['labels'])


    def create_model(self, lr=0.00001, optimizer='Adam', model=None):
        from keras.models import Model, load_model
        from keras.layers import Dense, Conv2D, Flatten, MaxPooling2D, Dropout, Input, concatenate
        from keras import regularizers
        from keras.optimizers import SGD, Adam     

        # if model then load existing trained model
        if model:
            # This is not right until we also save the training and testing sets
            model_path = pathlib.Path(model)
            logger.info('Using existing model: %s',model_path.resolve())
            self.class_model = load_model(model)

        # else, create a new model
        else:
            number_of_classes = len(self.label_list)

            inputs_A = Input(shape=(2,self.matrice_size,1))
            output_1A = Conv2D(32, kernel_size=5, activation='relu', padding='same', name='conv2d_1')(inputs_A)
            output_2A = MaxPooling2D(pool_size=(1, 2))(output_1A)
            output_3A = Conv2D(64, kernel_size=3, activation='relu', padding='same', name='conv2d_2')(output_2A)

            output_4A = Flatten(name='flatten_1')(output_3A)

            output_5A = Dense(128, kernel_regularizer=regularizers.l2(0.001), activation='relu', name='dense_1')(output_4A)
            output_6A = Dropout(0.5, name='dropout_1')(output_5A)
            prediction_class = Dense(number_of_classes, activation='softmax', name='dense_prediction')(output_6A)
            self.class_model = Model(inputs=inputs_A, outputs=prediction_class)            

            if optimizer != 'Adam':
                opt = SGD(lr=lr)
            else:
                opt = Adam(lr=lr)
            logger.info('Compiling model')
            self.class_model.compile(loss = "categorical_crossentropy", optimizer = opt, metrics=['accuracy','mae'])
        return None


    def get_model_summary(self):
        self.class_model.summary()


    def train_model(self, epochs=1000):
        self.class_model.fit(self.X_train, self.y_train, validation_data=(self.X_test, self.y_test), epochs=epochs)            


    def get_threshold(self, label='High_quality'):
        prob_df = self.get_true_vs_false_positive_df(label)
        prob_df['diff'] = prob_df['True'] - prob_df['False']
        threshold = prob_df['Probablity_threshold'][prob_df['diff'].idxmax()]
        return threshold


    def predict_peaks(self, threshold, samples=[], label='High_quality'):
        # Get the index of the label monitored
        probability_index = self.label_list.index(label)
        # If the threshold is not set manually
        # Get the otpimum threshold for label assignment
        # Make sure the sample list is not empty
        if not samples:
            samples = self.experiment.samples
        logger.info('Predicting peaks from {} samples'.format(len(samples)))
        # Create batches of data, one batch of data corresponds to all peaks in one sample
        for sample in samples:
            # Temporary fix to remove all existing formatted chromatogram created during training as they are pandas dataframes and not numpy arrays
            # TODO: rewrite the training part to use numpy arrays so this fix is not required anymore
            for peak in sample.peak_list:
                peak.formatted_chromatograms = []
                # Hotfix to add a prediction attribute to the peak object
                # TODO: Add it to the class definition
                peak.prediction = None

            # Generate all peak dataframe for the specific sample
            logger.info("Extracting and formatting peak chromatograms for sample: %s",sample.name)
            sample.create_interpolated_chromatograms(self.matrice_size, self.margin)
            data = []
            # Create the batch by adding all formatted chromatograms in one list
            valid_peak_list = []
            for peak in sample.peak_list:
                # Only add the peak if it is valid
                if peak.valid:
                    dataframe = peak.get_formatted_chromatogram(self.matrice_size, self.margin)
                    data.append([dataframe[0],dataframe[1]])
                    valid_peak_list.append(peak)

            # Transform the data batch (list) to a numpy array
            batch = np.asarray(data)
            # Reshape the batch so it can be fed to the Neural Network
            batch = batch.reshape(batch.shape[0],2,self.matrice_size,1).astype('float32')
            # Make prediction for the batch (all peaks present in this sample)
            logger.info('Predicting %d peaks',batch.shape[0])
            batch_prediction =  self.class_model.predict(batch)
            # Iterate through peaks to assign the predicted label
            for i in range(len(valid_peak_list)):
                # Test if the probability of the monitored label is above the threshold
                if batch_prediction[i][probability_index] >= threshold:
                    # If so set this label as the predicted peak label
                    self.annotation_table.set_peak_predicted_label(valid_peak_list[i], label)
                # Else, get the label associated with the prediction and assign it to the peak
                else:
                    # Get the index of the highest probability
                    predicted_arg = np.argmax(batch_prediction[i])
                    # Get the corresponding label
                    predicted_label = self.label_list[predicted_arg]
                    # Set this label as the predicted peak label
                    self.annotation_table.set_peak_predicted_label(valid_peak_list[i], predicted_label)


    def get_true_vs_false_positive_df(self, label='High_quality'):
        # Get the index of the label monitored
        probability_index = self.label_list.index(label)

        low_quality_index = self.label_list.index('Low_quality')
        # Create a dataframe to store all predicted values
        prob_array = np.arange(0,1.0,0.01)
        true_prediction = np.zeros(len(prob_array))
        false_prediction = np.zeros(len(prob_array))
        false_low = np.zeros(len(prob_array))
        false_noise = np.zeros(len(prob_array))

        prob_df = pd.DataFrame(list(zip(prob_array, true_prediction, false_prediction, false_low, false_noise)), 
                               columns =['Probablity_threshold', 'True', 'False', 'False_low', 'False_noise'])

        true_prediction = 0
        false_prediction = 0

        # Iterate through the test data and predict probability of belonging to the given label
        for i in range(len(self.X_validation)+1):
            j = i-1
            if j >= 0:
                prediciton_one_hot = self.class_model.predict(self.X_validation[j:i])

                # Get the real label of the peak
                real_arg = np.argmax(self.y_validation[j])
                # Get the predicted probability for the monitored label
                prediction_prob = prediciton_one_hot[0][probability_index]
                # If the label of the peak == the label monitor we have a true peak
                if real_arg == probability_index:
                    prob_df.loc[prob_df['Probablity_threshold'] <= prediction_prob, 'True'] += 1
                # Else this is a noise peak
                elif real_arg == low_quality_index:
                    prob_df.loc[prob_df['Probablity_threshold'] <= prediction_prob, 'False_low'] += 1
                    prob_df.loc[prob_df['Probablity_threshold'] <= prediction_prob, 'False'] += 1
                else:
                    prob_df.loc[prob_df['Probablity_threshold'] <= prediction_prob, 'False_noise'] += 1
                    prob_df.loc[prob_df['Probablity_threshold'] <= prediction_prob, 'False'] += 1

        prob_df['True'] = prob_df['True']/prob_df['True'].max()
        prob_df['False'] = prob_df['False']/prob_df['False'].max()
        prob_df['False_low'] = prob_df['False_low']/prob_df['False_low'].max()
        prob_df['False_noise'] = prob_df['False_noise']/prob_df['False_noise'].max()


        return prob_df




import pandas as pd
import pathlib

from .peak import Peak

class Feature():

    """
    Representation of an MS feature.

    Can contains several peaks that are grouped together from the same sample (adduct, isotopic peak, fragment...),
    mz and RT attributes correspond to the monoisotopic peak.
    Attached to a feature collection object that contain features from other samples aligned with this instance.  
    """

    # Feature counter to automatically assign feature ids
    feature_counter = 0
    feature_dict = dict()

    # This class method returns the feature instance from a feature id
    # This is to avoid iteration through all instances of features to find a specific one
    @classmethod
    def get_feature(cls, secondary_id):
        return cls.feature_dict[secondary_id]

    def __init__(self, mz, RT, sample=None, secondary_id=None, feature_collection=None):
        self.id = Feature.feature_counter
        self.secondary_id = secondary_id
        self.mz = mz
        self.RT = RT
        self.sample = sample
        self.peak_list = []
        self.feature_collection = feature_collection
        Feature.feature_counter += 1
        if secondary_id != None:
            Feature.feature_dict[self.secondary_id] = self


class FeatureCollection():

    """
    Representation of features from several samples aligned together

    Only contain a list of feature.
    """

    # Feature collection counter to automatically assign feature collection ids
    feature_collection_counter = 0

    def __init__(self):
         self.id = FeatureCollection.feature_collection_counter
         self.feature_list = []
         FeatureCollection.feature_collection_counter += 1


    def get_monoisotopic_peak_number(self, prediction_class, use_annotation):
        peak_count = 0
        for feature in self.feature_list:
            for peak in feature.peak_list:
                if peak.valid:
                    if (peak.monoisotopic is None) | (peak.monoisotopic is True):
                        if use_annotation:
                            if peak.annotation:
                                if peak.annotation.label in prediction_class:
                                    peak_count += 1
                            else:
                                if peak.prediction.label in prediction_class:
                                    peak_count += 1
                        else:
                            if peak.prediction.label in prediction_class:
                                peak_count += 1
        return peak_count


    def export_data(self, sample_list, export_classes, use_annotation, export_properties):
        # ["rt", "mz", "peak_mz", "peak_rt", "peak_rt_start", "peak_rt_end", "peak_mz_min", "peak_mz_max", "label", "height", "area"]
        data_dictionnary = dict()
        if "mz" in export_properties:
            data_dictionnary["m/z"] = self.feature_list[0].mz
        if "rt" in export_properties:
            data_dictionnary["retention time"] = self.feature_list[0].RT
        for feature in self.feature_list:
            sample_str = feature.sample.name
            for peak in feature.peak_list:
                if peak.valid:
                    if (peak.monoisotopic is None) | (peak.monoisotopic is True):
                        for export_property in export_properties:
                            if export_property == "peak_mz":
                                entry_str = ' '.join([sample_str,'Peak m/z'])
                                data_dictionnary[entry_str] = peak.mz
                            if export_property == "peak_rt":
                                entry_str = ' '.join([sample_str,'Peak RT'])
                                data_dictionnary[entry_str] = peak.RT
                            if export_property == "peak_rt_start":
                                entry_str = ' '.join([sample_str,'Peak RT start'])
                                data_dictionnary[entry_str] = peak.rt_start
                            if export_property == "peak_rt_end":
                                entry_str = ' '.join([sample_str,'Peak RT end'])
                                data_dictionnary[entry_str] = peak.rt_end
                            if export_property == "peak_mz_min":
                                entry_str = ' '.join([sample_str,'Peak m/z min'])
                                data_dictionnary[entry_str] = peak.mz_min
                            if export_property == "peak_mz_max":
                                entry_str = ' '.join([sample_str,'Peak m/z max'])
                                data_dictionnary[entry_str] = peak.mz_max
                            if export_property == "label":
                                entry_str = ' '.join([sample_str,'Peak label'])
                                data_dictionnary[entry_str] = peak.prediction.label
                                if use_annotation and peak.annotation:
                                    data_dictionnary[entry_str] = peak.annotation.label
                            if export_property == "height":
                                entry_str = ' '.join([sample_str,'Peak height'])
                                data_dictionnary[entry_str] = peak.height
                            if export_property == "area":
                                entry_str = ' '.join([sample_str,'Peak area'])
                                data_dictionnary[entry_str] = peak.area
        return data_dictionnary


class FeatureTable():

    """
    Representation of the feature extraction performed on the dataset

    Contains the feature table input file and enable its loading.
    Automate the creation of features and features object to the experiment.
    """

    # Feature table counter to automatically assign feature table ids
    feature_table_counter = 0

    def __init__(self, feature_table_path=None, origin=None):
        self.id = FeatureTable.feature_table_counter
        self.feature_table_path = feature_table_path
        self.feature_table = None
        self.column_map = None
        self.annotation_table = None
        self.feature_collection_list = []
        self.origin = origin
        FeatureTable.feature_table_counter += 1


    def load_feature_table(self):
        feature_table_file_folder = pathlib.Path(self.feature_table_path)
        if self.feature_table_path == None:
            print("Please, provdide path to a feature table")
            return None
        # If the path is a directory (Unaligned peaks, one table per sample)
        elif feature_table_file_folder.is_dir():
            # List all .csv files
            feature_table_files = [file for file in feature_table_file_folder.iterdir() if file.suffix.lower() in ['.csv']]
            # Create an empty dataframe
            feature_table = pd.DataFrame()
            # Iterate through all files (samples)
            for feature_table_file in feature_table_files:
                # Read the sample specific feature table
                sample_feature_table = pd.read_csv(feature_table_file)
                # Join the table to the main dataframe
                feature_table = feature_table.append(sample_feature_table, sort=False)
                # feature_table = pd.concat([feature_table, sample_feature_table], axis=1, sort=False)
            self.feature_table = feature_table.fillna(0)
            return feature_table
        else:
            feature_table = pd.read_csv(self.feature_table_path)
            self.feature_table = feature_table
            return feature_table


    def create_column_map(self, samples):
        feature_samples = []
        for column in self.feature_table.columns[2:]:
            feature_samples.append(column.split('.')[0])

        column_mapping = dict()
        for sample in samples:
            column_mapping[sample] = dict()

        for column in self.feature_table.columns[2:]:
            if 'Peak m/z' in column and column.split(' ')[-1] == 'm/z':
                column_mapping[column.split('.')[0]]['mz'] = column
            elif 'Peak RT' in column and column.split(' ')[-1] == 'RT':
                column_mapping[column.split('.')[0]]['RT'] = column
            elif 'Peak RT start' in column:
                column_mapping[column.split('.')[0]]['rt_start'] = column
            elif 'Peak RT end' in column:
                column_mapping[column.split('.')[0]]['rt_end'] = column
            elif 'Peak m/z min' in column:
                column_mapping[column.split('.')[0]]['mz_min'] = column
            elif 'Peak m/z max' in column:
                column_mapping[column.split('.')[0]]['mz_max'] = column
            elif 'Peak height' in column:
                column_mapping[column.split('.')[0]]['height'] = column
            elif 'Peak area' in column:
                column_mapping[column.split('.')[0]]['area'] = column
        self.column_map = column_mapping
        return column_mapping


    def load_features(self, sample_list):
        feature_number = self.feature_table.shape[0]
        i = 0
        # Iterate through all features in the feature table
        for index, row in self.feature_table.iterrows():
            i += 1
            # Extract feature specific information (mz and RT, RT is converted in seconds)
            mz = row['row m/z']
            RT = row['row retention time'] 
            feature_collection = FeatureCollection()
            # Iterate through all samples in the experiment
            for sample in sample_list:
                # Extract peak specific data for individual sample
                peak_dict = dict(
                    sample = sample,
                    RT = row[self.column_map[sample.name]['RT']], 
                    mz = row[self.column_map[sample.name]['mz']],
                    rt_start = row[self.column_map[sample.name]['rt_start']],
                    rt_end = row[self.column_map[sample.name]['rt_end']],
                    mz_min = row[self.column_map[sample.name]['mz_min']],
                    mz_max = row[self.column_map[sample.name]['mz_max']],
                    height = row[self.column_map[sample.name]['height']],
                    area = row[self.column_map[sample.name]['area']]
                )
                # We do not save the peak if it does not exist in this sample (i.e. RT == 0 in table)
                if peak_dict['RT'] != 0:
                    # Create feature (No secondary id)
                    feature = Feature(mz, RT, sample, feature_collection=feature_collection)
                    # Create peak
                    new_peak = Peak(sample, peak_dict['RT'], peak_dict['mz'], peak_dict['rt_start'], peak_dict['rt_end'], peak_dict['mz_min'], 
                        peak_dict['mz_max'], peak_dict['height'], peak_dict['area'])
                    # Add feature to the peak (one peak is attached to a unique feature)
                    new_peak.feature = feature
                    # Add the peak to the sample
                    sample.peak_list.append(new_peak)
                    # Add the peak to the feature
                    feature.peak_list.append(new_peak)
                    # Add the feature to the sample
                    sample.feature_list.append(feature)
                    # Add the feature to the feature collection (Only one feature per feature collection in this case)
                    feature_collection.feature_list.append(feature)
            # Add the feature to the feature list (all peaks belonging to this feature have now been created)
            self.feature_collection_list.append(feature_collection)
        return None




import random

class Annotation():

    """
    Representation of a label


    Contains a list of peaks which have been assigned a specific label
    """

    # Annotation counter to automatically assign annotation ids
    annotation_counter = 0

    def __init__(self, label):
        self.id = Annotation.annotation_counter
        self.label = label
        self.peaks = []
        Annotation.annotation_counter += 1


class AnnotationTable():

    """
    Keep track of peaks that have been labelled.

    Contains ordered list labelled peaks and unlabelled peaks.
    Facilitate labelling of new peaks. 
    """

    # Annotation table counter to automatically assign annotation table ids
    annotation_table_counter = 0

    def __init__(self, feature_table=None, labels=[]):
        self.id = AnnotationTable.annotation_table_counter
        self.feature_table = feature_table
        self.labelled_peaks = []
        self.unlabelled_peaks = self.create_unlabelled_peak_list()
        self.annotations = self.create_annotation_objects(labels)
        AnnotationTable.annotation_table_counter += 1

    def create_annotation_objects(self, labels):
        annotations = []
        for label in labels:
            annotation = Annotation(label)
            annotations.append(annotation)
        return annotations

    def create_unlabelled_peak_list(self, monoisotopic_only=False):
        if self.feature_table:
            # unlabelled_peaks = [peak for feature in self.feature_table.feature_list for peak in feature.peak_list]
            unlabelled_peaks = []
            for feature_collection in self.feature_table.feature_collection_list:
                for feature in feature_collection.feature_list:
                    for peak in feature.peak_list:
                        if monoisotopic_only: 
                            if peak.monoisotopic:
                                unlabelled_peaks.append(peak)
                        else:
                            unlabelled_peaks.append(peak)
            self.unlabelled_peaks = unlabelled_peaks
            return unlabelled_peaks
        else:
            print("No feature table yet attached to the annotation table")
            return None

    def get_peak_to_label(self):
        random_peak = random.choice(self.unlabelled_peaks)
        return random_peak

    def get_annotation(self, label):
        for annotation in self.annotations:
            if annotation.label == label:
                return annotation
        print("No annotation with the label exist!")
        return None

    def set_peak_label(self, peak, label):
        # If the peak already has an annotation
        if peak.annotation:
            # Remove the peak from the peak list in the annotation object
            peak.annotation.peaks.remove(peak)
            # We do not remove the annotation from the peak as it will be replaced by the new one
        annotation = self.get_annotation(label)
        peak.annotation = annotation
        annotation.peaks.append(peak)
        # If the peak is not already in the labelled peak list
        if peak not in self.labelled_peaks:
            # Add the peak
            self.labelled_peaks.append(peak)
        # If the peak is still in the unlabelled peak list
        if peak in self.unlabelled_peaks:
            # Remove the peak
            self.unlabelled_peaks.remove(peak) 


    def set_peak_predicted_label(self, peak, label):
        # We do not touch peak annotation here, only prediction
        # If the peak already has a prediction
        if peak.prediction:
            # Remove the peak from the peak list in the annotation (prediction) object
            peak.preditcion.peaks.remove(peak)
            # We do not remove the prediction from the peak as it will be replaced by the new one
        # Get the new annotation object using the label 
        prediction = self.get_annotation(label)
        # Assign it to the peak as prediction
        peak.prediction = prediction
        # Add the peak to the annotation object peak list
        prediction.peaks.append(peak)


class AnnotationTool():

    """
    The annotation tool only points to an annotation table. 
    The tool can access the already labelled peaks using the ordered list in the annotation table.
    New peaks to annotate are always selected randomly from the unlabelled peak list.

    This class implements the jupyter notebook integrated interface for peak labelling.

    Extra external libraries are required to launch the annotation tool. See documentation for  more details. 
    """

    def __init__(self, experiment, margin=1, review=[], table_index=0):


        self.annotation_table = experiment.feature_tables[table_index].annotation_table
        # Retention time window margins
        self.margin = margin
        # List of classes to review, if the list is not empty, review mode is considered 'ON'
        self.review = review

    def launch_annotation_tool(self):
        from jupyter_plotly_dash import JupyterDash

        import dash
        import dash_core_components as dcc
        import dash_html_components as html
        import plotly.graph_objs as go
        # Initialise with the first peak to annotate
        # Every variable has to be assigned to self to be accessible by the annotation tool due to function closure
        # Get a random peak to annotate
        if not self.review:
            self.peak = self.annotation_table.get_peak_to_label()
            # Set a variable to track the "previous" button click (go through the labelled peak list backwards)
            # When the buttin next is clicked:
            # If previous == 0, the peak to labelled will be randomly selected from the unlabelled peak list
            # If previous < 0, we return the corresponding peak in the labelled peak list
            # Previous is initialised to 0 as the first peak is randomly selected from the unlabelled peak list
            self.previous = 0
        else:
            # When reviewing peaks, we set previous to the first peak in the labelled peak list
            self.previous = -len(self.annotation_table.labelled_peaks)
            review_peak = False
            # We increment previous until we find a peak that is labelled with a class in the review list 
            while review_peak == False:
                # If the peak label is in the review list
                if self.annotation_table.labelled_peaks[self.previous].annotation.label in self.review:
                    # Assign this peak to the current peak 
                    self.peak = self.annotation_table.labelled_peaks[self.previous]
                    # Set review_peak to True to get out of the loop
                    review_peak = True
                else:
                    # Otherwise we increment previous to check the next peak
                    self.previous += 1

        # Get the list of possible labels regardless of review mode
        self.labels = [annotation.label for annotation in self.annotation_table.annotations]
        self.label_options = []
        for label in self.labels:
            self.label_options.append(dict(label=label.replace('_',' '), value=label))

        # This is currently not supported when Dash is embedded in Jupyter notebook
        external_stylesheets = ['https://codepen.io/chriddyp/pen/bWLwgP.css']

        # app = dash.Dash(__name__, external_stylesheets=external_stylesheets)
        app = JupyterDash('NeatMS_annotation_tool')

        app.layout = html.Div(children=[
            # Removed all useless default title and messages TODO: decide if we want to add anything
            # html.H1(children='Peak classification app'),

            # html.Div(children='''
            #     Dash: A web application framework for Python.
            # '''),

            dcc.Graph(
                id='peak-graph',
            ),
            html.Div([
                html.Div([
                    dcc.RadioItems(
                        id='input-class',
                        options=self.label_options,
                        value='Noise',
                        labelStyle={'display': 'inline-block', 'text-align': 'justify', 'margin-right': '2%'} #THIS IS WHERE IT IS ADDED
                    ),
                ], style={'margin-bottom':'2%','textAlign': 'center'}),
                html.Div([
                    html.Button('Previous', id='btn-previous', n_clicks_timestamp=0),
                    html.Button('Next', id='btn-next', n_clicks_timestamp=0),
                ], style={'margin-bottom':'2%','textAlign': 'center'}),
                html.Div([
                    dcc.Slider(
                        id='slider-updatemargin',
                        min=0,
                        max=1,
                        step=0.05,
                        value=self.margin,
                        marks={
                            0: {'label': '0'},
                            0.25: {'label': '0.25'},
                            0.50: {'label': '0.50'},
                            0.75: {'label': '0.75'},
                            1: {'label': '1'}
                        },
                        updatemode='drag'
                    ),
                    html.Div(id='updatemargin-output-container', style={'margin-top': 20}) 
                ], style={'width': '100%', 'textAlign': 'center', 'justify-content': 'center'}),
            ], style={'textAlign': 'center'})

        ])

        def create_line_plot(chromatogram, boundaries):
            return {
                'data': [
                        go.Line(x=chromatogram[0], y=chromatogram[1])
                    ],
                'layout': dict(
                        title= str('Peak %i - m/z: %.4f (%.4f-%.4f)' % (boundaries['id'], boundaries['mz'], boundaries['mz_min'], boundaries['mz_max'])),
                        shapes = [
                            dict(
                                type = 'line',
                                x0 = boundaries['rt_start'],
                                y0 = 0,
                                x1 = boundaries['rt_start'],
                                y1 = boundaries['height'],
                                line = dict(
                                    color = 'rgb(0, 0, 0)',
                                    width = 2
                                )
                            ),
                            dict(
                                type = 'line',
                                x0 = boundaries['rt_end'],
                                y0 = 0,
                                x1 = boundaries['rt_end'],
                                y1 = boundaries['height'],
                                line = dict(
                                    color = 'rgb(0, 0, 0)',
                                    width = 2
                                )
                            )
                        ],
                        xaxis = dict(
                            title = 'Retention time (min)'
                        ),
                        yaxis = dict(
                             title = 'Intensity'
                        )
                )

            }

        @app.callback(dash.dependencies.Output('updatemargin-output-container', 'children'),
              [dash.dependencies.Input('slider-updatemargin', 'value')])
        def display_value(value):
            return 'Retention time window margin: {:0.0f}%'.format(value*100)

        @app.callback(
            dash.dependencies.Output('peak-graph', 'figure'),
                    [dash.dependencies.Input('btn-next', 'n_clicks_timestamp'),
                    dash.dependencies.Input('btn-previous', 'n_clicks_timestamp'),
                    dash.dependencies.Input('slider-updatemargin', 'value')],
                    [dash.dependencies.State('input-class', 'value')])
        def updatePeak(next_peak, previous_peak, margin_value, class_value):
            # If the margin value has changed, no button has been click, only the slider was changed
            if margin_value != self.margin:
                # We only update the margin value
                self.margin = margin_value
            else:
                # If next or previous button has been clicked
                if next_peak > 0 or previous_peak > 0:
                    # If next button has been clicked
                    if next_peak > previous_peak:
                        # If we are looking at peaks already annotated 
                        if self.previous < -1:
                            # Change annotation of the previous peak
                            self.annotation_table.set_peak_label(self.peak, class_value)
                            # If review mode is 'OFF'
                            if not self.review:
                                # Increment the previous counter
                                self.previous += 1
                                # Get the next peak in the list
                                self.peak = self.annotation_table.labelled_peaks[self.previous]
                            # If review mode is 'ON'
                            else:
                                review_peak = False
                                self.previous += 1
                                # We increment previous until we find a peak that is labelled with a class in the review list 
                                while review_peak == False:
                                    # If the peak label is in the review list
                                    if self.annotation_table.labelled_peaks[self.previous].annotation.label in self.review:
                                        # Assign this peak to the current peak 
                                        self.peak = self.annotation_table.labelled_peaks[self.previous]
                                        # Set review_peak to True to get out of the loop
                                        review_peak = True
                                    else:
                                        # Otherwise we increment previous to check the next peak
                                        self.previous += 1
                        else:
                            # Change annotation of the previous peak (Noise for now)
                            self.annotation_table.set_peak_label(self.peak, class_value)
                            # If we are looking at unlabelled peaks, make sure the previous counter is 0
                            self.previous = 0
                            # Get a random peak
                            self.peak = self.annotation_table.get_peak_to_label()
                    # If previous button was clicked
                    else :
                        # Change annotation of the previous peak (Noise for now)
                        self.annotation_table.set_peak_label(self.peak, class_value)
                        # If first click on previous, decrement by 2 as we just added the new labelled peak as the last peak in the list
                        if self.previous == 0:
                            self.previous -= 2
                        # Else we just decrement by 1
                        else:
                            # Decrement the previous counter 
                            self.previous -= 1
                        # If review mode is 'OFF'
                        if not self.review:
                            # Get the already annotated peak from the list
                            self.peak = self.annotation_table.labelled_peaks[self.previous]
                        # If review mode is 'ON'
                        else:
                            # Decrement of previous already applied above
                            review_peak = False
                            # We decrement previous until we find a peak that is labelled with a class in the review list 
                            while review_peak == False:
                                # If the peak label is in the review list
                                if self.annotation_table.labelled_peaks[self.previous].annotation.label in self.review:
                                    # Assign this peak to the current peak 
                                    self.peak = self.annotation_table.labelled_peaks[self.previous]
                                    # Set review_peak to True to get out of the loop
                                    review_peak = True
                                else:
                                    # Otherwise we increment previous to check the next peak
                                    self.previous -= 1
                else:
                    # Initialisation, no button has been yet clicked, returning the initial random peak
                    self.peak = self.peak
            chromatogram = self.peak.get_chromatogram(self.margin)
            boundaries = dict(
                id = self.peak.id,
                rt_start = self.peak.rt_start,
                rt_end = self.peak.rt_end,
                height = self.peak.get_height(),
                mz_min = self.peak.mz_min,
                mz_max = self.peak.mz_max,
                mz = self.peak.mz
                )
            return create_line_plot(chromatogram, boundaries)

        @app.callback(
            dash.dependencies.Output('input-class', 'value'),
                    [dash.dependencies.Input('peak-graph', 'figure')])
        def updatePeak(data):
            if self.peak.annotation != None:
                class_value = self.peak.annotation.label
            else:
                class_value = 'Noise'
            return class_value

        return app




# -*- coding: utf-8 -*-

"""
NeatMS library
~~~~~~~~~~~~~~

NeatMS is a peak filtering library for Mass Spectrometry data.

The code is open-source and available at

Full documentation is available at

:licence: MIT, see LICENCE for more details.

"""


import pymzml
import numpy as np
import pandas as pd
import pathlib
import time
import random
import pickle
import sys

from .experiment import Experiment
from .data import RawData, DataReader, PymzmlDataReader #, OpenmsDataReader
from .feature import Feature, FeatureCollection, FeatureTable
from .annotation import Annotation, AnnotationTable, AnnotationTool
from .peak import Chromatogram, Peak
from .handler import NN_handler
from .sample import Sample


# Set default logging handler to avoid "No handler found" warnings.
import logging
from logging import NullHandler

logging.getLogger(__name__).addHandler(NullHandler())


from distutils.core import setup
setup(
  name = 'NeatMS',
  packages = ['NeatMS'],
  version = 'v0.5',
  license='MIT', 
  description = 'NeatMS is an open source python package for untargeted LCMS signal labelling and filtering', 
  author = 'Yoann Gloaguen', 
  author_email = 'yoann.gloaguen@mdc-berlin.de', 
  url = 'https://github.com/bihealth/NeatMS', 
  download_url = 'https://github.com/bihealth/NeatMS/archive/v0.5.tar.gz',
  keywords = ['LCMS', 'Classifier', 'Peak', 'Neural network'], 
  install_requires=[ 
          'pymzml',
          'numpy',
          'pandas',
          'scikit-learn',
          'tensorflow',
          'pillow',
          'h5py',
          'keras'
      ],
  classifiers=[
    'Development Status :: 4 - Beta',
    'Intended Audience :: Science/Research',   
    'Topic :: Scientific/Engineering :: Bio-Informatics',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7'
  ],
)